# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

"""Model helper to handle sort by and sort direction"""

from enum import Enum


class CaseInsensitiveEnum(Enum):
    """Enum overriden to support case insensitive keys"""

    @classmethod
    def _missing_(cls, value):
        for member in cls:
            if member.value == value.upper():
                return member


class SortBy(CaseInsensitiveEnum):
    """Values to filter data by"""

    RUNTIME = "runtime"
    STATUS = "status"
    STARTED = "started_at"
    LATTICE_NAME = "lattice_name"
    ENDED = "ended_at"


class SortDirection(CaseInsensitiveEnum):
    """Values to decide sort direction"""

    ASCENDING = "ASC"
    DESCENDING = "DESC"



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

""" Monitor executor instances."""


from typing import Any

from covalent._shared_files import logger
from covalent.executor.base import _AbstractBaseExecutor as _ABE
from covalent.executor.utils import Signals

from ..data_modules import job_manager

app_log = logger.app_log
log_stack_info = logger.log_stack_info

_putters = {}
_getters = {}


async def _get_cancel_requested(dispatch_id: str, task_id: int):
    """
    Query the database for the task's cancellation status

    Arg(s)
        dispatch_id: Dispatch ID of the lattice
        task_id: ID of the task within the lattice

    Return(s)
        Cancellation status of the task

    """
    # Don't hit the DB for post-processing task
    if task_id < 0:
        return False

    app_log.debug(f"Get _handle_requested for executor {dispatch_id}:{task_id}")
    job_records = await job_manager.get_jobs_metadata(dispatch_id, [task_id])
    app_log.debug(f"Job record: {job_records[0]}")
    return job_records[0]["cancel_requested"]


async def _put_job_handle(dispatch_id: str, task_id: int, job_handle: str) -> bool:
    """
    Store the job handle of the task returned by the backend in the database

    Arg(s)
        dispatch_id: Dispatch ID of the lattice
        task_id: ID of the task within the lattice
        job_handle: Unique identifier of the task returned by the execution backend

    Return(s)
        True
    """
    # Don't hit the DB for post-processing task
    if task_id < 0:
        return False
    app_log.debug(f"Put job_handle for executor {dispatch_id}:{task_id}")
    await job_manager.set_job_handle(dispatch_id, task_id, job_handle)
    return True


_putters["job_handle"] = _put_job_handle
_getters["cancel_requested"] = _get_cancel_requested


async def _handle_message(
    dispatch_id: str, task_id: int, executor: _ABE, action: Signals, body: Any = None
) -> Any:
    """
    Handle the action properly

    Arg(s)
        dispatch_id: Dispatch ID of the lattice
        task_id: ID of the task in the lattice
        executor: Instance of the abstract base executor
        action: Action requested to be performed
        body: Content of the action/request made

    Return(s)
        Response corresponding to the action requested
    """
    if action == Signals.GET:
        return await _getters[body](dispatch_id, task_id)

    if action == Signals.PUT:
        key, val = body
        await _putters[key](dispatch_id, task_id, val)
        return None
    else:
        raise KeyError(f"Unknown action {action}")


async def watch(dispatch_id: str, task_id: int, executor: _ABE) -> None:
    """
    Watch the send and receive queues of the executor

    Arg(s)
        dispatch_id: Dispatch ID of the lattice
        task_id: ID of the task within the lattice
        executor: Instance of the abstract base executor

    Return(s)
        None
    """
    send_queue = executor._send_queue
    recv_queue = executor._recv_queue

    while True:
        action, body = await send_queue.get()
        app_log.debug(f"Received message {action} from executor {dispatch_id}:{task_id}")

        if action == Signals.EXIT:
            app_log.debug(f"Stopping listener for executor {dispatch_id}:{task_id}")
            break

        try:
            resp = await _handle_message(dispatch_id, task_id, executor, action, body)
            recv_queue.put_nowait((True, resp))
        except Exception as ex:
            app_log.warning(f"Error handling message {action} from executor: {ex}")
            recv_queue.put_nowait((False, None))
            break



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

"""
Defines the core functionality of the result service
"""

import asyncio
import traceback
import uuid
from datetime import datetime, timezone
from typing import Callable, Dict, Optional

from covalent._results_manager import Result
from covalent._shared_files import logger
from covalent._shared_files.defaults import sublattice_prefix
from covalent._shared_files.util_classes import RESULT_STATUS
from covalent._workflow.lattice import Lattice
from covalent._workflow.transport_graph_ops import TransportGraphOps

from .._db import load, update, upsert
from .._db.write_result_to_db import resolve_electron_id

app_log = logger.app_log
log_stack_info = logger.log_stack_info

# References to result objects of live dispatches
_registered_dispatches = {}

# Map of dispatch_id -> message_queue for pushing node status updates
# to dispatcher
_dispatch_status_queues = {}


def generate_node_result(
    node_id: int,
    node_name: str,
    start_time=None,
    end_time=None,
    status=None,
    output=None,
    error=None,
    stdout=None,
    stderr=None,
    sub_dispatch_id=None,
    sublattice_result=None,
):
    """
    Helper routine to prepare the node result

    Arg(s)
        node_id: ID of the node in the transport graph
        node_name: Name of the node
        start_time: Start time of the node
        end_time: Time at which the node finished executing
        status: Status of the node's execution
        output: Output of the node
        error: Error from the node
        stdout: STDOUT of a node
        stderr: STDERR generated during node execution
        sub_dispatch_id: Dispatch ID of the sublattice
        sublattice_result: Result of the sublattice

    Return(s)
        Dictionary of the inputs
    """
    return {
        "node_id": node_id,
        "node_name": node_name,
        "start_time": start_time,
        "end_time": end_time,
        "status": status,
        "output": output,
        "error": error,
        "stdout": stdout,
        "stderr": stderr,
        "sub_dispatch_id": sub_dispatch_id,
        "sublattice_result": sublattice_result,
    }


async def _handle_built_sublattice(dispatch_id: str, node_result: Dict) -> None:
    """Make dispatch for sublattice node.

    Note: The status COMPLETED which invokes this function refers to the graph being built. Once this step is completed, the sublattice is ready to be dispatched. Hence, the status is changed to DISPATCHING.

    Args:
        dispatch_id: Dispatch ID
        node_result: Node result dictionary

    """
    try:
        node_result["status"] = RESULT_STATUS.DISPATCHING_SUBLATTICE
        result_object = get_result_object(dispatch_id)
        sub_dispatch_id = await make_sublattice_dispatch(result_object, node_result)
        node_result["sub_dispatch_id"] = sub_dispatch_id
        node_result["start_time"] = datetime.now(timezone.utc)
        node_result["end_time"] = None
    except Exception as ex:
        tb = "".join(traceback.TracebackException.from_exception(ex).format())
        node_result["status"] = RESULT_STATUS.FAILED
        node_result["error"] = tb
        app_log.debug(f"Failed to make sublattice dispatch: {tb}")


# Domain: result
async def update_node_result(result_object, node_result) -> None:
    """
    Updates the result object with the current node_result

    Arg(s)
        result_object: Result object the current dispatch
        node_result: Result of the node to be updated in the result object

    Return(s)
        None

    """
    app_log.debug(f"Updating node result for {node_result['node_id']}.")

    if (
        node_result["status"] == RESULT_STATUS.COMPLETED
        and node_result["node_name"].startswith(sublattice_prefix)
        and not node_result["sub_dispatch_id"]
    ):
        app_log.debug(
            f"Sublattice {node_result['node_name']} build graph completed, invoking make sublattice dispatch..."
        )
        await _handle_built_sublattice(result_object.dispatch_id, node_result)

    try:
        update._node(result_object, **node_result)
    except Exception as ex:
        app_log.exception(f"Error persisting node update: {ex}")
        node_result["status"] = RESULT_STATUS.FAILED
    finally:
        sub_dispatch_id = node_result["sub_dispatch_id"]
        detail = {"sub_dispatch_id": sub_dispatch_id} if sub_dispatch_id is not None else {}
        if node_status := node_result["status"]:
            dispatch_id = result_object.dispatch_id
            status_queue = get_status_queue(dispatch_id)
            node_id = node_result["node_id"]
            await status_queue.put((node_id, node_status, detail))


# Domain: result
def initialize_result_object(
    json_lattice: str, parent_result_object: Result = None, parent_electron_id: int = None
) -> Result:
    """Convenience function for constructing a result object from a json-serialized lattice.

    Args:
        json_lattice: a JSON-serialized lattice
        parent_result_object: the parent result object if json_lattice is a sublattice
        parent_electron_id: the DB id of the parent electron (for sublattices)

    Returns:
        Result: result object

    """
    dispatch_id = get_unique_id()
    lattice = Lattice.deserialize_from_json(json_lattice)
    result_object = Result(lattice, dispatch_id)
    if parent_result_object:
        result_object._root_dispatch_id = parent_result_object._root_dispatch_id

    result_object._electron_id = parent_electron_id
    result_object._initialize_nodes()
    app_log.debug("2: Constructed result object and initialized nodes.")

    update.persist(result_object, electron_id=parent_electron_id)
    app_log.debug("Result object persisted.")

    return result_object


# Domain: result
def get_unique_id() -> str:
    """
    Get a unique ID.

    Args:
        None

    Returns:
        str: Unique ID

    """
    return str(uuid.uuid4())


async def make_dispatch(
    json_lattice: str, parent_result_object: Result = None, parent_electron_id: int = None
) -> str:
    """Make a dispatch from a json-serialized lattice.

    Args:
        json_lattice: a JSON-serialized lattice.
        parent_result_object: the parent result object if json_lattice is a sublattice.
        parent_electron_id: the DB id of the parent electron (for sublattices).

    Returns:
        Dispatch ID of the lattice.

    """
    result_object = initialize_result_object(
        json_lattice, parent_result_object, parent_electron_id
    )
    _register_result_object(result_object)
    return result_object.dispatch_id


async def make_sublattice_dispatch(result_object: Result, node_result: dict) -> str:
    """Get sublattice json lattice (once the transport graph has been built) and invoke make_dispatch.

    Args:
        result_object: Result object for parent dispatch of the node.
        node_result: Result of the node.

    Returns:
        str: Dispatch ID of the sublattice.

    """
    node_id = node_result["node_id"]
    json_lattice = node_result["output"].object_string
    parent_electron_id = load.electron_record(result_object.dispatch_id, node_id)["id"]
    app_log.debug(
        f"Making sublattice dispatch for node_id {node_id} and electron_id {parent_electron_id}."
    )
    return await make_dispatch(json_lattice, result_object, parent_electron_id)


def _get_result_object_from_new_lattice(
    json_lattice: str, old_result_object: Result, reuse_previous_results: bool
) -> Result:
    """Get new result object for re-dispatching from new lattice json.

    Args:
        json_lattice: JSON-serialized lattice.
        old_result_object: Result object of the previous dispatch.

    Returns:
        Result object.

    """
    lat = Lattice.deserialize_from_json(json_lattice)
    result_object = Result(lat, get_unique_id())
    result_object._initialize_nodes()

    if reuse_previous_results:
        tg = result_object.lattice.transport_graph
        tg_old = old_result_object.lattice.transport_graph
        reusable_nodes = TransportGraphOps(tg_old).get_reusable_nodes(tg)
        TransportGraphOps(tg).copy_nodes_from(tg_old, reusable_nodes)

    return result_object


def _get_result_object_from_old_result(
    old_result_object: Result, reuse_previous_results: bool
) -> Result:
    """Get new result object for re-dispatching from old result object.

    Args:
        old_result_object: Result object of the previous dispatch.
        reuse_previous_results: Whether to reuse previous results.

    Returns:
        Result: Result object for the new dispatch.

    """
    result_object = Result(old_result_object.lattice, get_unique_id())
    result_object._num_nodes = old_result_object._num_nodes

    if not reuse_previous_results:
        result_object._initialize_nodes()

    return result_object


def make_derived_dispatch(
    parent_dispatch_id: str,
    json_lattice: Optional[str] = None,
    electron_updates: Optional[Dict[str, Callable]] = None,
    reuse_previous_results: bool = False,
) -> str:
    """Make a re-dispatch from a previous dispatch.

    Args:
        parent_dispatch_id: Dispatch ID of the parent dispatch.
        json_lattice: JSON-serialized lattice of the new dispatch.
        electron_updates: Dictionary of electron updates.
        reuse_previous_results: Whether to reuse previous results.

    Returns:
        str: Dispatch ID of the new dispatch.

    """
    if electron_updates is None:
        electron_updates = {}

    old_result_object = load.get_result_object_from_storage(parent_dispatch_id)

    if json_lattice:
        result_object = _get_result_object_from_new_lattice(
            json_lattice, old_result_object, reuse_previous_results
        )
    else:
        result_object = _get_result_object_from_old_result(
            old_result_object, reuse_previous_results
        )

    result_object.lattice.transport_graph.apply_electron_updates(electron_updates)
    result_object.lattice.transport_graph.dirty_nodes = list(
        result_object.lattice.transport_graph._graph.nodes
    )
    update.persist(result_object)
    _register_result_object(result_object)
    app_log.debug(f"Redispatch result object: {result_object}")

    return result_object.dispatch_id


def get_result_object(dispatch_id: str) -> Result:
    return _registered_dispatches[dispatch_id]


def _register_result_object(result_object: Result):
    dispatch_id = result_object.dispatch_id
    _registered_dispatches[dispatch_id] = result_object
    _dispatch_status_queues[dispatch_id] = asyncio.Queue()


def finalize_dispatch(dispatch_id: str):
    del _dispatch_status_queues[dispatch_id]
    del _registered_dispatches[dispatch_id]


def get_status_queue(dispatch_id: str):
    return _dispatch_status_queues[dispatch_id]


async def persist_result(dispatch_id: str):
    result_object = get_result_object(dispatch_id)
    update.persist(result_object)
    await _update_parent_electron(result_object)


async def _update_parent_electron(result_object: Result):
    if parent_eid := result_object._electron_id:
        dispatch_id, node_id = resolve_electron_id(parent_eid)
        status = result_object.status
        if status == RESULT_STATUS.POSTPROCESSING_FAILED:
            status = RESULT_STATUS.FAILED
        parent_result_obj = get_result_object(dispatch_id)
        node_result = generate_node_result(
            node_id=node_id,
            node_name=parent_result_obj.lattice.transport_graph.get_node_value(node_id, "name"),
            end_time=result_object.end_time,
            status=status,
            output=result_object._result,
            error=result_object._error,
            sub_dispatch_id=load.sublattice_dispatch_id(parent_eid),
            sublattice_result=result_object,
        )

        app_log.debug(f"Updating sublattice parent node {dispatch_id}:{node_id}")
        await update_node_result(parent_result_obj, node_result)


def upsert_lattice_data(dispatch_id: str):
    result_object = get_result_object(dispatch_id)
    upsert.lattice_data(result_object)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

import json
from copy import deepcopy
from functools import wraps
from typing import Callable, Dict, List, Optional, Union

import requests

from .._results_manager import wait
from .._results_manager.result import Result
from .._results_manager.results_manager import get_result
from .._shared_files import logger
from .._shared_files.config import get_config
from .._workflow.lattice import Lattice
from ..triggers import BaseTrigger
from .base import BaseDispatcher

app_log = logger.app_log
log_stack_info = logger.log_stack_info


def get_redispatch_request_body(
    dispatch_id: str,
    new_args: Optional[List] = None,
    new_kwargs: Optional[Dict] = None,
    replace_electrons: Optional[Dict[str, Callable]] = None,
    reuse_previous_results: bool = False,
) -> Dict:
    """Get request body for re-dispatching a workflow."""
    if new_args is None:
        new_args = []
    if new_kwargs is None:
        new_kwargs = {}
    if replace_electrons is None:
        replace_electrons = {}
    if new_args or new_kwargs:
        res = get_result(dispatch_id)
        lat = res.lattice
        lat.build_graph(*new_args, **new_kwargs)
        json_lattice = lat.serialize_to_json()
    else:
        json_lattice = None
    updates = {k: v.electron_object.as_transportable_dict for k, v in replace_electrons.items()}

    return {
        "json_lattice": json_lattice,
        "dispatch_id": dispatch_id,
        "electron_updates": updates,
        "reuse_previous_results": reuse_previous_results,
    }


class LocalDispatcher(BaseDispatcher):
    """
    Local dispatcher which sends the workflow to the locally running
    dispatcher server.
    """

    @staticmethod
    def dispatch(
        orig_lattice: Lattice,
        dispatcher_addr: str = None,
        disable_run: bool = False,
    ) -> Callable:
        """
        Wrapping the dispatching functionality to allow input passing
        and server address specification.

        Afterwards, send the lattice to the dispatcher server and return
        the assigned dispatch id.

        Args:
            orig_lattice: The lattice/workflow to send to the dispatcher server.
            dispatcher_addr: The address of the dispatcher server.  If None then defaults to the address set in Covalent's config.
            disable_run: Whether to disable running the workflow and rather just save it on Covalent's server for later execution

        Returns:
            Wrapper function which takes the inputs of the workflow as arguments
        """

        if dispatcher_addr is None:
            dispatcher_addr = (
                "http://"
                + get_config("dispatcher.address")
                + ":"
                + str(get_config("dispatcher.port"))
            )

        @wraps(orig_lattice)
        def wrapper(*args, **kwargs) -> str:
            """
            Send the lattice to the dispatcher server and return
            the assigned dispatch id.

            Args:
                *args: The inputs of the workflow.
                **kwargs: The keyword arguments of the workflow.

            Returns:
                The dispatch id of the workflow.
            """

            # To access the disable_run passed to the dispatch function
            nonlocal disable_run

            lattice = deepcopy(orig_lattice)

            lattice.build_graph(*args, **kwargs)

            # Serialize the transport graph to JSON
            json_lattice = lattice.serialize_to_json()

            # Extract triggers here
            json_lattice = json.loads(json_lattice)
            triggers_data = json_lattice["metadata"].pop("triggers")

            if not disable_run:
                # Determine whether to disable first run based on trigger_data
                disable_run = triggers_data is not None

            json_lattice = json.dumps(json_lattice)

            submit_dispatch_url = f"{dispatcher_addr}/api/submit"

            r = requests.post(
                submit_dispatch_url, data=json_lattice, params={"disable_run": disable_run}
            )
            r.raise_for_status()

            lattice_dispatch_id = r.content.decode("utf-8").strip().replace('"', "")

            if not disable_run or triggers_data is None:
                return lattice_dispatch_id

            LocalDispatcher.register_triggers(triggers_data, lattice_dispatch_id)

            return lattice_dispatch_id

        return wrapper

    @staticmethod
    def dispatch_sync(
        lattice: Lattice,
        dispatcher_addr: str = None,
    ) -> Callable:
        """
        Wrapping the synchronous dispatching functionality to allow input
        passing and server address specification.

        Afterwards, sends the lattice to the dispatcher server and return
        the result of the executed workflow.

        Args:
            orig_lattice: The lattice/workflow to send to the dispatcher server.
            dispatcher_addr: The address of the dispatcher server. If None then defaults to the address set in Covalent's config.

        Returns:
            Wrapper function which takes the inputs of the workflow as arguments.
        """

        if dispatcher_addr is None:
            dispatcher_addr = (
                "http://"
                + get_config("dispatcher.address")
                + ":"
                + str(get_config("dispatcher.port"))
            )

        @wraps(lattice)
        def wrapper(*args, **kwargs) -> Result:
            """
            Send the lattice to the dispatcher server and return
            the result of the executed workflow.

            Args:
                *args: The inputs of the workflow.
                **kwargs: The keyword arguments of the workflow.

            Returns:
                The result of the executed workflow.
            """

            return get_result(
                LocalDispatcher.dispatch(lattice, dispatcher_addr)(*args, **kwargs),
                wait=wait.EXTREME,
            )

        return wrapper

    @staticmethod
    def redispatch(
        dispatch_id: str,
        dispatcher_addr: str = None,
        replace_electrons: Dict[str, Callable] = None,
        reuse_previous_results: bool = False,
        is_pending: bool = False,
    ) -> Callable:
        """
        Wrapping the dispatching functionality to allow input passing and server address specification.

        Args:
            dispatch_id: The dispatch id of the workflow to re-dispatch.
            dispatcher_addr: The address of the dispatcher server. If None then then defaults to the address set in Covalent's config.
            replace_electrons: A dictionary of electron names and the new electron to replace them with.
            reuse_previous_results: Boolean value whether to reuse the results from the previous dispatch.

        Returns:
            Wrapper function which takes the inputs of the workflow as arguments.
        """

        if dispatcher_addr is None:
            dispatcher_addr = (
                "http://"
                + get_config("dispatcher.address")
                + ":"
                + str(get_config("dispatcher.port"))
            )

        if replace_electrons is None:
            replace_electrons = {}

        def func(*new_args, **new_kwargs):
            """
            Prepare the redispatch request body and redispatch the workflow.

            Args:
                *args: The inputs of the workflow.
                **kwargs: The keyword arguments of the workflow.

            Returns:
                The result of the executed workflow.

            """
            body = get_redispatch_request_body(
                dispatch_id, new_args, new_kwargs, replace_electrons, reuse_previous_results
            )
            redispatch_url = f"{dispatcher_addr}/api/redispatch"
            r = requests.post(redispatch_url, json=body, params={"is_pending": is_pending})
            r.raise_for_status()
            return r.content.decode("utf-8").strip().replace('"', "")

        return func

    @staticmethod
    def register_triggers(triggers_data: List[Dict], dispatch_id: str) -> None:
        """
        Register the given triggers to the Triggers server.
        Register also starts the `observe()` method of said trigger.
        This is done by calling `BaseTrigger._register` method
        with the given trigger dictionary and is equivalent to
        calling the trigger objects `register()` method.

        Args:
            triggers_data: List of trigger dictionaries to be registered
            dispatch_id: Lattice's dispatch id to be linked with given triggers

        Returns:
            None
        """

        for tr_dict in triggers_data:
            tr_dict["lattice_dispatch_id"] = dispatch_id
            BaseTrigger._register(tr_dict)

    @staticmethod
    def stop_triggers(
        dispatch_ids: Union[str, List[str]], triggers_server_addr: str = None
    ) -> None:
        """
        Stop observing on all triggers of all given dispatch ids registered on the Triggers server.
        Args:
            dispatch_ids: Dispatch ID(s) for whose triggers are to be stopped
            triggers_server_addr: Address of the Triggers server; configured dispatcher's address is used as default
        Returns:
            None
        """

        if triggers_server_addr is None:
            triggers_server_addr = (
                "http://"
                + get_config("dispatcher.address")
                + ":"
                + str(get_config("dispatcher.port"))
            )

        stop_triggers_url = f"{triggers_server_addr}/api/triggers/stop_observe"

        if isinstance(dispatch_ids, str):
            dispatch_ids = [dispatch_ids]

        r = requests.post(stop_triggers_url, json=dispatch_ids)
        r.raise_for_status()

        app_log.debug("Triggers for following dispatch_ids have stopped observing:")
        for d_id in dispatch_ids:
            app_log.debug(d_id)
